;###############################################################3
; Textual interface


(define menu-choice
  (lambda ()
    (let ((c (read)))
    (cond ((eqv? c 1) (Load-Spec)(menu))
          ((eqv? c 2) (Load-Web)(menu))
          ((eqv? c 3) (debug web specification)(newline)(display "Press Return to continue")(read-char)(read-char)(menu))
          ((eqv? c 4)(newline)(newline)
                     (display "-------------------- Web site ----------------------")
                     (Show-Web readable-web)(newline)
                     (display "----------------------------------------------------")(newline)
                     (display "Press Return to continue")(read-char)(read-char)(menu))
          ((eqv? c 5)(newline)(newline)
                     (display "---------------- Web specification ------------------")(newline)(newline)
                     (Show-Spec readable-spec)
                     (display "----------------------------------------------------")(newline)
                     (display "Press Return to continue")(read-char)(read-char)(menu))
          ((eqv? c 6) (display "Excellent!"))
          (else (menu))
          )
      )
    )
  )
  
(define menu
  (lambda ()
    (newline)
    (display "**********************************")(newline)
    (display "              M E N U")(newline)
    (display "**********************************")(newline)
    (display "1. Load web specification")
    (newline)   
    (display "2. Load web site")
    (newline)
    (display "3. Check web site")
    (newline)
    (display "4. Show current web site")
    (newline)
    (display "5. Show current web specification")
    (newline)
    (display "6. Quit")
    (newline)
    (display "**********************************")(newline)
    (newline)        
    (display "Choose: ")
    (menu-choice)
    )
  )

(define main 
 (lambda ()
   (display "*******************************************")(newline)
   (display "      VERDI: VErification and Rewriting for Debugging Internet Sites  v0.2   ")(newline)
   (display "*******************************************")(newline)(newline)
   (menu)
   )
  )
      
(define Show-Web
  (lambda (readable-web)
    (if (not (null? readable-web))
        (begin
          (display (car readable-web))(newline)(newline)
          (Show-Web (cdr readable-web))
          )
        )
    )
 )

(define Show-Spec
  (lambda (readable-spec)
    (if (not (null? readable-spec))
        (begin
          (display (car readable-spec))(newline)(newline)
          (Show-Spec (cdr readable-spec))
          )
        )
    )
 ) 
; The web specification is a list ((lhs1,rhs1), (lhs2,rhs2),...)
; stored in the global variable Specification

(define Specification '())

;Load-spec: load a web specification from a file into the global variable Specification

(define Load-Spec
  (lambda ()
    (newline)
    (set! readable-spec '())
    (display "Web specification file name? ")
    (set! Specification (Parse-Spec (open-input-file (symbol->string (read)))))
    (newline)(display "Web Specification loaded!")(newline)
    )
  )

; The web is stored in the global list Web is a list (term1,term2,...)
(define Web '())

; Load-Web: loads a web site into the global variable Web.

(define Load-Web
  (lambda ()
    (newline)
    (set! readable-web '())
    (display "Web site file name? ")
    (set! Web (Parse-Web (open-input-file (symbol->string (read)))))
    (newline)(display "Web Site loaded!")(newline)
    )
  )

;Get-Page: gets a page from a web site file 
; (format of the file: semistructured expression <newline>%<newline> semistructured expression ...)
;input: an input port
;output: a string containing a semistructured expression

(define Get-Page
   (lambda (inport)
       (let ((r (read-line inport))) 
         (if  (or (eof-object? r) (string=? r "%")) 
              ""
              (string-append r (Get-Page inport)))
         )
     )
  )           
 
;Parse-Web: parses a web site and builds its internal representation
;input: an input port 
;output: a list of terms (i.e. web pages)

(define Parse-Web
    (lambda (inport)
            (let ((page (Get-Page inport)))
              (if (not (eof-object? page)) (set! readable-web (cons page readable-web)))
              (if (string=? "" page)
                  '()
                  (cons (car (Term (translate-token-list(scan-expression page)))) (Parse-Web inport)) 
             )
        )
      )
  )


;Parse-Sec: parses a web specification and builds an internal representation
;input: an input port 
;output: a list of program rules

(define Parse-Spec
    (lambda (inport)
      (let ((r (read-line inport)))
        (if (not (eof-object? r)) (set! readable-spec (cons r readable-spec)))
        (if  (eof-object? r) 
             '()
             (cons (Rule (token-list r)) (Parse-Spec inport)) 
             )
        )
      )
  )

(define readable-web '())
(define readable-spec '())

;##########################################################################
;  an LL(1) parser (with a very weak error-handling)
;  for the following grammar

; Rule  ::=  Tag "(" Termlist ")"  "->" Term
; Term ::= Variable | Text | Tag "(" Termlist ")"
; Termlist ::= Term Rest | Rest
; Rest ::= "," Term | Rest | empty

(define string-suffix
   (lambda (s n)
     (if (< n (string-length s))
         (if (or (or (char=? (string-ref s n) #\space) (char=? (string-ref s n) #\newline)) 
                 (char=? (string-ref s n) #\tab)) 
             (string-suffix s (+ n 1))
             (substring s n (string-length s))
             )
      ""  
      )
     )
  )
 

; A very simple lexer...
       
(define token-list
  (lambda (s)
           (let* ((token (regexp"[#]?[a-zA-Z][a-zA-Z0-9]*|[#]?['][a-zA-Z0-9]*'|\\(|\\)|,|\\->")) 
                  (first  (regexp-match token s)))
             (if first (append first (token-list (string-suffix s (string-length (car first)) )))
                 ()
                 )
             )
    )
)  

(define isVar?
   (lambda (token)
      (if (or (char-upper-case? (string-ref token 0)) (and (char=? #\# (string-ref token 0)) 
                                                           (char-upper-case? (string-ref token 1)))) #t #f)
     )
  )
     

(define isTag?
   (lambda (token)
      (if (or (char-lower-case? (string-ref token 0)) (and (char=? #\# (string-ref token 0)) 
                                                           (char-lower-case? (string-ref token 1)))) #t #f)
     )
  )

(define isText?
   (lambda (token)
      (if (or (char=? (string-ref token 0) #\')  (and (char=? #\# (string-ref token 0)) 
                                                      (char=? #\' (string-ref token 1)))) #t #f)
     )
  )

(define shift
  (lambda (tk-list token)
    (if (null? tk-list) 'PARSE-ERROR
        (if (string=? (car tk-list) token)
            (cdr tk-list)
            'PARSE-ERROR
            )
        )
    )
  )
  
(define  Rule
   (lambda (tk-list)
     (if (or (eqv? tk-list 'PARSE-ERROR) (null? tk-list))
         'PARSE-ERROR
         (let ((lookahead (car tk-list))) 
           (if  (isTag? lookahead)
                (let* ((out (Term tk-list)) (remaining-input  (shift (cadr out) "->")))
                  (if (or (eqv? out 'PARSE-ERROR) (eqv? remaining-input 'PARSE-ERROR))
                      'PARSE-ERROR
                      (list  (car out) (car (Term remaining-input)))
                      )
                  )
                'PARSE-ERROR
                )
           )
         )
     )
  )

(define Term
  (lambda (tk-list)
    (if (eqv? tk-list 'PARSE-ERROR)
        'PARSE-ERROR
        (let ((lookahead (car tk-list)))
          (cond ((isVar? lookahead) (cons  (list lookahead) (list (cdr tk-list))))
                ((isText? lookahead)  (cons (list lookahead) (list (cdr tk-list))))
                ((isTag? lookahead) (let ((out (Term-List (shift (cdr tk-list) "("))))
                                      (if (or (eqv? out 'PARSE-ERROR) (eqv? (shift (cadr out) ")") 'PARSE-ERROR))
                                          'PARSE-ERROR
                                          (cons (list lookahead (car out)) (list (cdr (cadr out))))
                                          )
                                      )
                 )                                         
                (else 'PARSE-ERROR)
                )
          )
        )
    )
)

(define Term-List
   (lambda (tk-list)
     (if (or (null? tk-list)(eqv? tk-list 'PARSE-ERROR))
         'PARSE-ERROR
         (let ((lookahead (car tk-list)))
           (if (string=? lookahead ")")   
               (Rest '() tk-list)
               (letrec (( t (Term tk-list))) ; (inerithed, remaining tokens)
                 (if (eqv? t 'PARSE-ERROR)
                     'PARSE-ERROR
                     (Rest (car t) (cadr t))
                     )
                 )
               )
           )
         )
     )
  )
  

(define Rest
   (lambda (inherited-list tk-list)
     (if (eqv? tk-list 'PARSE-ERROR)
         'PARSE-ERROR
         (letrec ((lookahead (car tk-list)))
           (cond ((string=? lookahead ",") 
                  (letrec (( t (Term (shift tk-list ",")))) (Rest (append inherited-list (car t)) (cadr t))))
                 ((string=? lookahead ")") (cons inherited-list (list tk-list)))
                 (else 'PARSE-ERROR)
           )
         )
     )
  )
)
 
;###################################################################################

;translation of semistructured expressions into terms

       
(define scan-expression
  (lambda (s)
           (let* ((token (regexp"[<][/a-zA-Z][a-zA-Z0-9]*[>]|[a-zA-Z0-9]+")) 
                  (first  (regexp-match token s)))
             (if first (append first (scan-expression (string-suffix s (string-length (car first)) )))
                 ()
                 )
             )
    )
)  

(define translate-token-list
    (lambda (tk-list)
      (if (null? tk-list) '()
       (let ((first (car tk-list)))
         (cond 
           ((not (char=? (string-ref first 0) #\<)) 
            (cons (string-append (string-append (string #\') first) (string #\')) (translate-token-list (cdr tk-list))))
           ((and (char=? (string-ref first 0) #\<) (not (char=? (string-ref first 1) #\/))) 
            (append (list (substring first 1 (- (string-length first) 1)) (string #\()) (translate-token-list (cdr tk-list))))
           ((and (char=? (string-ref first 0) #\<)  (char=? (string-ref first 1) #\/))
            (if (not(null? (cdr tk-list)))
                (let ((second (cadr tk-list)))
                  (if (and (char=? (string-ref second 0) #\<) (char=? (string-ref second 1) #\/))
                      (cons ")"   (translate-token-list (cdr tk-list)))
                      (append (list  ")" ",")  (translate-token-list (cdr tk-list)))
                ))
                (cons ")"   (translate-token-list (cdr tk-list)))
                )
            )
           (else 'TRANSLATION-ERROR)
            )
         )
       )
      )
  )

;###############################################################################3
; simulation, partial matching and all that stuff




(define arity
  (lambda (term)
     (if (null? term) 0
         (let ((first (car term)))
         (if (isTag? first) 
             (+ 1 (arity (cddr term)))
             (+ 1 (arity (cdr term)))
             )
           )
         )
    )
  )


(define adj-list '())
(define next-pos 0)

; computes the preorder of a tree and the edges of the tree

(define previsit 
  (lambda (t pos)
    (set! next-pos (+ next-pos 1))
    (if (null? t) '()
        (let ((first (car t)))
         (cond ((or (or (isVar? first) (isText? first)) (and (isTag? first) (null? (cadr t)))) 
                (set! adj-list (cons (cons pos 'NULL) adj-list)) 
                (list first))
               (else (cons (car t) (previsit-aux (cadr t) pos))) 
               )
          )
        )
    )
  )
   

(define lessthan
  (lambda (p1 p2)
    (if (<= (car p1) (car p2)) #t #f)
    )
  )
    
(define previsit-aux
  (lambda (args pos)
    (if (null? args) '()
        (let ((first (car args)))
          (set! adj-list (cons (cons pos next-pos) adj-list)) 
          (if (isTag? first) 
              (append (previsit (list first (cadr args)) next-pos) (previsit-aux (cddr args) pos))
              (append (previsit (list first) next-pos) (previsit-aux (cdr args) pos))
              )
          )
        )
    )
  )

(define format-adj-elem
  (lambda (adj-list)
    (if (null? adj-list) '()
        (if (or (null? (cdr adj-list)) (not (= (caar adj-list) (caadr adj-list))))
            (list (car adj-list))
            (cons (car adj-list) (format-adj-elem (cdr adj-list)))
            )
        )
    )
  )

(define format-adj-list
  (lambda (adj-list)
    (if (null? adj-list) '()
        (let ((first (format-adj-elem adj-list)))
          (cons (format-adj-elem adj-list) (format-adj-list (list-tail adj-list (length first))))
          )
        )
    )
  )
    

(define make-adj-list
  (lambda (t)
    (let  ((labels (previsit t 0)) (aux  (quicksort adj-list lessthan)))
      (set! next-pos 0)(set! adj-list '())
      (let* ((nedges (length aux)) (n (length labels)) (parent (make-vector n)))
        (vector-set! parent 0 0)
        (let loop ((i 0))
          (if (< i nedges)
              (let ((edge (list-ref aux i)))
                (if (not(eqv? 'NULL (cdr edge)))
                    (vector-set! parent (cdr edge) (car edge))
                    )
                (loop (+ i 1))
                )
              )
          )
        (list labels parent (format-adj-list aux))
        )
      )
    )
  )

    

(define cur-pos -1)

(define select  
  (lambda (t pos)
    (set! cur-pos (+ cur-pos 1))
    (let ((first (car t)))
      (cond ((= pos cur-pos)  (set! cur-pos -1) t)
            ((or (or (isVar? first) (isText? first)) (and (isTag? first) (null? (cadr t))))  #f)
            (else (select-aux (cadr t) pos)) 
        )
      )
    )
  )
    

(define select-aux
  (lambda (args pos)
    (if (null? args) #f
        (let ((first (car args)))
          (if (isTag? first) 
                   (let*  ((firstarg (list first (cadr args))) (infirstarg? (select firstarg pos))) 
                            (if infirstarg? infirstarg? 
                                (select-aux (cddr args) pos))
                     )
                   (let*  ((firstarg (list  first)) (infirstarg? (select firstarg pos))) 
                     (if infirstarg? infirstarg? 
                         (select-aux (cdr args) pos))
                     )
            )
          )
        )
    )
  )
        

(define mark=?
  (lambda (r1 r2)
    (cond  ((char=? (string-ref r1 0) #\#) (mark=? (substring r1 1 (string-length r1)) r2))
              ((char=? (string-ref r2 0) #\#) (mark=? r1 (substring r2 1 (string-length r2))))
              (else (if (string=? r1 r2) #t #f))
              )
    )
  )

(define label-relation? 
 (lambda (i label1 parent-label1 j label2 parent-label2)
   (if (isVar? label1)
       (if (= i 0) 
           #t
           (if (not (= j 0))
               (mark=? parent-label1 parent-label2)
               #f
               )
           )
       (mark=? label1 label2)
       )
   )
  )
        
(define simulation
  (lambda (t1 t2)
    (let* ((tinfo1 (make-adj-list t1)) (tinfo2 (make-adj-list t2)) (nodes1 (car tinfo1)) (nodes2 (car tinfo2)) (n1 (length nodes1)) 
           (n2 (length nodes2)) (sim (make-vector n1))) 
      (let loop1 ((i 0))
        (if (< i n1)
            (begin
              (vector-set! sim i '())
              (let loop2 ((j 0))
                          (if (< j n2)
                              (begin
                                (if (label-relation? i (list-ref nodes1 i) (list-ref nodes1 (vector-ref (cadr tinfo1) i)) 
                                                     j (list-ref nodes2 j)(list-ref nodes2 (vector-ref (cadr tinfo2) j))) 
                                    (vector-set! sim i (cons (cons i j) (vector-ref sim i))))
                                (loop2 (+ 1 j))
                                )
                              )
                          )
                (loop1 (+ 1 i))       
                )
              )
            )
      (let  ((sim-aux (vector->list (remove-couple sim tinfo1 tinfo2))))
        (if (member '() sim-aux) #f
            (substitutions (filter-all (injective sim-aux) (cadr tinfo1) (cadr tinfo2)) t1 t2 nodes1 nodes2)
            )
        )
        )
      )
    )

(define project2
  (lambda (l)
    (if (null? l) '()
        (map cdr l)
        )
    )
  )

(define remove-couple 
  (lambda (sim tinfo1 tinfo2)
    (let ((nsim (vector-length sim)) (adjlist1 (list-ref tinfo1 2)) (adjlist2 (list-ref tinfo2 2)) (nodes1 (list-ref tinfo1 0)) 
          (nodes2 (list-ref tinfo2 0)))
       (let loop1 ((i 0))
         (if (< i nsim)
             (let* ((sim_i (vector-ref sim i)) (nsim_i (length sim_i)))
               (let loop2 ((j 0))
                 (if (< j nsim_i)
                     (let  ((couple (list-ref sim_i j)))
                       (if  (not (label-test (project2 (list-ref adjlist1 (car couple))) 
                                       (project2 (list-ref adjlist2 (cdr couple))) nodes1 nodes2))
                            (vector-set! sim (car couple) (remove couple (vector-ref sim (car couple))))
                           )
                       (loop2 (+ j 1))
                       )
                     )
                 )
               (loop1 (+ i 1))
               )
             )
         )
      sim
      )
    )
  )
             
(define label-test
  (lambda (adj_i adj_j label_1 label_2)
    (if (eq? 'NULL (car adj_i)) #t
        (let* ((l_i (lambda (n) (if (eq? 'NULL n) "nomatch1" (list-ref label_1 n))))
              (l_j (lambda (n) (if (eq? 'NULL n) "nomatch2" (list-ref label_2 n)))) (ni (length adj_i))
              (lab_i (map l_i adj_i)) (lab_j (map l_j adj_j)))
          (let loop ((k 0))
            (if (< k ni)
                (cond  ((isVar? (list-ref lab_i k)) (if (eq? 'NULL (car adj_j)) #f (loop (+ k 1)))) 
                       (else (if (membermark (list-ref lab_i k) lab_j) 
                                 (loop (+ k 1))
                                 #f
                                 )
                             )
                       )
                #t
                )
            )
          )
        )
    )
  )
                
(define membermark
  (lambda (x l)
    (if (null? l) #f
        (if (mark=? x (car l)) #t
            (membermark x (cdr l))
            )
        )
    )
  )


(define injective
  (lambda (sim)
    (if (null? sim) '()
        (cond ((null? (cdr sim)) (map list (car sim)))
              (else 
               (let* ((aux (injective (cdr sim))) (Sols '()) (first (car sim)) (n (length first)))
                 (let loop ((k 0))
                   (if (< k n)
                       (begin
                         (set! Sols (append (map (lambda (S) (cons (list-ref first k) S)) aux) Sols))
                         (loop (+ k 1))
                         )
                       )
                   )
                 Sols
                 )         
               )
              )
        )
    )
  )


(define filter-all 
  (lambda (inj-sim parent-vect1 parent-vect2)
    (if (null? inj-sim) '()
        (let ((fl (filter-single (car inj-sim) parent-vect1 parent-vect2)))
          (if (< (length fl) (vector-length parent-vect1))
              (filter-all (cdr inj-sim) parent-vect1 parent-vect2)
              (cons fl (filter-all (cdr inj-sim) parent-vect1 parent-vect2))
              )
          )
        )
    )
  )

 (define filter-single
   (lambda (inj-single parent-vect1 parent-vect2)
     (let ((filtered-list '()) (aux inj-single))
       (let loop ((inj-single inj-single))
         (if (not (null? inj-single)) 
             (let ((first (car inj-single)))
               (if (= (car first) 0)
                   (set! filtered-list (cons first filtered-list))
                   (if (member (cons (vector-ref parent-vect1 (car first)) (vector-ref parent-vect2 (cdr first))) aux)
                         (set! filtered-list (cons first filtered-list))
                         (set! filtered-list '())
                         )
                     )
               (loop (cdr inj-single))
               )
             )
         )
       filtered-list
       )
     )
   )
                             
     
(define substitutions
  (lambda (sim-list pattern page labels1 labels2)
      (if (null? sim-list) 
          '() 
          (cons (single-sub (car sim-list) pattern page labels1 labels2) (substitutions (cdr sim-list) pattern page labels1 labels2))
          )
    )
  )
  
    
 
(define single-sub
  (lambda (single-sim pattern page labels1 labels2)
    (if (null? single-sim) '()
        (let* ((first (car single-sim)) (li (list-ref labels1 (car first))))
          (if (isVar? li)
              ;(if (char=? #\# (string-ref li 0))
               ;   (cons (cons (unmarkvar li) (list (markterm (select page (cdr first))))) (single-sub (cdr single-sim) pattern page labels1 labels2))
                  (cons (cons (unmarkvar li) (list (select page (cdr first)))) (single-sub (cdr single-sim) pattern page labels1 labels2))
                ;  )
              (single-sub (cdr single-sim) pattern page labels1 labels2)
              )
          )
        )
    )
  )

;input: a (marked) variable (a string)
;output: a variable (it removes markings, if any)

(define unmarkvar
  (lambda (v)
    (if (char=? #\# (string-ref v 0)) (substring v 1 (string-length v)) v
        )
    )
  )

(define markterm
  (lambda (t)
    (if (null? t) '()
        (let ((first (car t)))
          (if (char=? #\# (string-ref first 0))  
              (cond  ((or (isVar? first) (isText? first)) (list first))
                     (else (list first (markterm-aux (cadr t))))
                     )
              (cond ((or (isVar? first) (isText? first)) (list (string-append "#" (car t))))
                    (else (list (string-append "#" (car t)) (markterm-aux (cadr t))))
                    )
              )
          )
        )
    )
  )
   
    
(define markterm-aux
  (lambda (args)
    (if (null? args) '()
        (let ((first (car args))) 
          (if (isTag? first) 
              (append (markterm (list first (cadr args))) (markterm-aux (cddr args)))
              (append (markterm (list first)) (markterm-aux (cdr args)))
              )
          )
        )
    )
  )  


; input: term t, rule r
; output: a set of term t' s.t. t ->r t'

(define partial-rewrite
  (lambda (t r)
    (let ((lhs (car  r)) (rhs (cadr r)) (rsigmas '()))
      (let loop ((subs (simulation lhs t)))
        (if (and subs (not (null? subs)))
                 (let ((aux (substitute rhs (car subs))))
              (if (not (member aux rsigmas))
                  (set! rsigmas (cons (substitute rhs (car subs)) rsigmas))
                  )
              (loop (cdr subs))
              )
            )
        )
      rsigmas
      )
    )
  )

; input: set of terms Tset, specification Spec
; output: set of terms Rset s.t. t' in Rset, t->Spec t' for some t in Tset

(define operatorT
  (lambda (Tset Spec)
    (let ((Rset '()))
      (let ((Tl (length Tset)) (Sl (length Spec)))
        (let loop1 ((i 0))
          (if (< i Tl)
              (begin
                (let loop2 ((j 0))
                  (if  (< j Sl)
                       (begin
                         (set! Rset (union Rset (partial-rewrite (list-ref Tset i) (list-ref Spec j))))
                         (loop2 (+ j 1))
                         )
                       )
                  )
                (loop1 (+ i 1))
                )
              )
          )
        )
      Rset
      )
    )
  )
  

;input: set set1, set set2 
;output set set1 union set2

(define union
 (lambda (set1 set2)
   (if (null? set1) set2
       (let ((first (car set1)))
         (if (not(member first set2))
             (cons first (union (cdr set1) set2))
             (union (cdr set1) set2)
             )
         )
       )
   )
 )
   
 
 
 
;input:  term t, substitution sigma
;output: term sigma(t)
   
(define substitute
  (lambda (t sigma)
      (if (null? t) '()
          (let ((first (car t)))  
              (cond  ((isVar? first) 
                      (let ((s (assoc (unmarkvar first) sigma)))
                        (cond (s (if (char=? #\# (string-ref first 0)) (markterm (cadr s)) (cadr s)))
                              (else t))
                        )
                      )
                     ((isText? first) t)
                     ((and (isTag? first) (null? (cadr t))) t)
                     (else (list first (substitute-aux (cadr t) sigma)))
                     )
            )
          )
    )
  )
   
    
(define substitute-aux
  (lambda (args sigma)
    (if (null? args) '()
        (let ((first (car args))) 
          (if (isTag? first) 
              (append (substitute (list first (cadr args)) sigma) (substitute-aux (cddr args) sigma))
              (append (substitute (list first) sigma) (substitute-aux (cdr args) sigma))
              )
          )
        )
    )
  )     

;input: a set of terms Tset, specification Spec
;output: set of Terms aux (in the paper, Req(Tset, Spec)) 

(define fixpointT
  (lambda (Tset Spec)
    (let ((pred '()) (aux (operatorT Tset Spec)))
      (let loop ((aux aux))
        (if (> (length aux) (length pred))
            (begin
              (set! pred aux)
              (set! aux (union (operatorT aux Spec) aux))
              (loop aux)
              )
            )
        aux
        )
      )
    )
  )
  

;input:  term t
;output: a term which is the marked part of term t
   
(define markpart
  (lambda (t)
      (if (null? t) '()
          (let ((first (car t)))  
              (cond  ((isVar? first) (if (char=? #\# (string-ref first 0)) t '()))
                     ((isText? first) (if (char=? #\# (string-ref first 0)) t '()))
                     ((and (isTag? first) (null? (cadr t))) (if (char=? #\# (string-ref first 0)) t '())) 
                     (else  (if (char=? #\# (string-ref first 0)) (list first (markpart-aux (cadr t))) '()))
                     )
            )
          )
    )
  )
   
    
(define markpart-aux
  (lambda (args)
    (if (null? args) '()
        (let ((first (car args))) 
          (if (isTag? first) 
              (append (markpart (list first (cadr args))) (markpart-aux (cddr args)))
              (append (markpart (list first)) (markpart-aux (cdr args)))
              )
          )
        )
    )
  ) 

;input a set of terms Tset, a requirement req
;output the set Cand(req)

(define cand
  (lambda (Tset req)
    (if (null? Tset) '() 
        (let ((first (car Tset)))
          (if (simulation (markpart req) first) 
              (cons first (cand (cdr Tset) (markpart req))) 
              (cand (cdr Tset) (markpart req)))
          )
        )
    )
  )
 
;input: set of Terms T, requirement req
;output: incompleteness/missing errors for req w.r.t T

(define errors-req   
    (lambda (T req)
      (let ((candidates (cand T req)))
        (if (null? candidates) (begin (display "Warning: ")(display req)(display " is missing!")(newline)(newline))
            (missing-err req candidates)
            )
            )
      )
  )

(define missing-err
  (lambda (req candidates)
    (if (not (null? candidates))
        (let ((first (car candidates)))
          (if (not (simulation req first))
              (begin
                (display "Warning: page ") 
                (display first) 
                (display " is incomplete w.r.t. ")(display "requirement ") (display req)(newline)(newline)
                )
              )
          (missing-err req (cdr candidates))
          )
        )
    )
  )

;input: set of terms Web, requirement set Req
;output: display incomplete/missing errors found in Web w.r.t Req

(define errors-all
  (lambda (web Req)
    (if (not(null? Req))
        (begin 
               (errors-req web (car Req)) (errors-all web (cdr Req)))
        )
    )
  )

;input: set of terms Web, specification Spec
;output: display incomplete/missing errors found in Web w.r.t Spec

(define debug 
  (lambda (Web Spec)
    (newline)(newline)
    (display "--------------- Errors found ----------------------")
    (newline)(newline)
    (errors-all Web (fixpointT Web Spec))
    (display "---------------------------------------------------")
    )
  )

; Main program starts here!

(main)

(define page (car web))
(define rule (car specification))
(define pattern (caar specification))
;###############################
  